import hashlib
import httplib
import random
import socket
import struct

from socket.web.exceptions import WebSocketException


SEC_KEY_RANDOM_CHARS = range(0x21, 0x2f + 1) + range(0x3a, 0x7e + 1)


def create_connection(address, timeout=None, **extra):
    socket = WebSocket()
    if timeout:
        socket.settimeout(timeout)
    socket.connect(address, **extra)
    return socket


class WebSocket(object):
    def __init__(self, secure=True):
        self.connected = False
        self.socket = socket.socket()
        self.secure = secure

    def settimeout(self, timeout):
        self.socket.settimeout(timeout)

    def gettimeout(self):
        return self.socket.gettimeout()

    def connect(self, address, **extra):
        host, port, location = address
        self.socket.connect((host, port))
        self.handshake(host, port, location, **extra)

    def handshake(self, host, port, location, **extra):
        hostname = "{0}:{1}".format(host, port) if port != 80 else host

        headers = {}
        headers["Upgrade"]= "WebSocket"
        headers["Connection"]= "Upgrade"
        headers["Host"] = hostname
        headers["Origin"] = hostname
        headers.update(extra)

        key1, number1 = self.generate_key()
        headers["Sec-WebSocket-Key1"] = key1
        key2, number2 = self.generate_key()
        headers["Sec-WebSocket-Key2"] = key2
        
        # Shuffle headers
        fields = [(key, value)
            for key, value in headers.iteritems()]
        random.shuffle(fields)
        headers = dict(fields)

        key3 = self.generate_third_key()
        
        http = httplib.HTTPConnection(host, port)
        http.sock = self.socket
        http.request("GET", location, key3, headers)
        response = http.getresponse()
        if response.status != 101:
            self.socket.close()
            raise WebSocketException(response.reason)
        if response.getheader("upgrade", "") != "WebSocket":
            if response.getheader("connection", "").lower() != "upgrade":
                self.socket.close()
                raise WebSocketException("Missing or bad upgrade")
        if response.getheader("sec-websocket-origin", "") != hostname:
            self.socket.close()
            raise WebSocketException("Missing or bad WebSocket-Origin")
        if response.getheader("sec-websocket-location", "") != "ws://{0}{1}".format(hostname, location):
            self.socket.close()
            raise WebSocketException("Missing or bad Sec-Websocket-Location")
        if self.secure and not self.validate_response(number1, number2, key3, response):
            self.socket.close()
            raise WebSocketException("Mismatch challenge/response")
        self.connected = True

    def validate_response(self, number1, number2, key3, response):
        challenge = struct.pack("!I", number1)
        challenge += struct.pack("!I", number2)
        challenge += key3
        digest = hashlib.md5(challenge).digest()
        return response.read(16) == digest

    def generate_key(self):
        """
        Generates handshake key as described in
        http://www.whatwg.org/specs/web-socket-protocol/
        """
        def uint32(i):
            return i % 2 ** 32
        spaces = random.randint(1, 12)
        maximum = int(4294967295 / uint32(spaces))
        number = uint32(random.randint(1, maximum + 1))
        product = number * uint32(spaces)
        key = "{0:d}".format(product)
        for i in range(random.randint(1, 12)):
            pos = random.randint(1, len(key))
            char = SEC_KEY_RANDOM_CHARS[random.randint(1, len(SEC_KEY_RANDOM_CHARS) - 1)]
            key = key[0:pos] + chr(char) + key[pos:]
        for i in range(spaces):
            pos = random.randint(1, len(key) - 1)
            key = key[0:pos] + " " + key[pos:]
        return key, number

    def generate_third_key(self):
        """
        Generates handshake key3 as described in
        http://www.whatwg.org/specs/web-socket-protocol/
        """
        return "".join([chr(random.randint(0, (1<<8) -1)) for i in range(8)])

    def send(self, data):
        if isinstance(data, unicode):
            data = data.encode("utf-8")
        data = "".join(["\x00", data, "\xff"])
        return self.socket.send(data)

    def recv(self):
        byte = self.socket.recv(1)
        if not byte:
            raise WebSocketException("Connection closed")
        frame = ord(byte)
        if frame == 0x00:
            bytes = []
            while True:
                byte = self.socket.recv(1)
                if byte == "\xff":
                    break
                else:
                    bytes.append(b)
            return "".join(bytes)
        elif 0x80 < frame < 0xff:
            length = 0
            while True:
                b = ord(self.socket.recv(1))
                length = length * (1 << 7) + (b & 0x7f)
                if b < 0x80:
                    break
            bytes = self.socket.recv(length)
            return bytes
        elif frame_type == 0xff:
            self.socket.recv(1)
            self.close()
            return None
        else:
            raise WebSocketException("Invalid frame type")

    def close(self):
        self.socket.close()
        self.connected = False
